@@ -69,7 +69,7 @@
 //!   agent.fetch_root_key().await?;
 //!   let management_canister_id = Principal::from_text("aaaaa-aa")?;
 //!
-//!   let waiter = delay::Delay::builder()
+//!   let waiter = garcon::Delay::builder()
 //!     .throttle(std::time::Duration::from_millis(500))
 //!     .timeout(std::time::Duration::from_secs(60 * 5))
 //!     .build();
@@ -120,4 +120,4 @@ pub use agent::{agent_error, agent_error::AgentError, nonce::NonceFactory, Agent
 pub use identity::{Identity, Signature};
 pub use request_id::{to_request_id, RequestId, RequestIdError};
 
-pub(crate) mod hash_tree;
+pub(crate) use ic_types::hash_tree;
